#!/usr/bin/env python3

import atexit
import os

from base64 import b64decode

from csp_client import *


logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]: %(message)s')
config = {}


def populate_config():
    global config

    # Global
    config["manifestType"] = os.getenv("manifestType", None)
    config["destroyMode"] = os.getenv("destroyMode", "False") in ["True", "true", "Yes", "yes"]
    config["cspBootstrapAPIToken"] = os.getenv("cspBootstrapAPIToken", None)
    config["cspBootstrapAPITokenCSR"] = os.getenv("cspBootstrapAPITokenCSR", None)
    config["cspUser"] = os.getenv("cspUser", None)
    config["cspPassword"] = os.getenv("cspPassword", None)
    config["cspUrl"] = os.getenv("cspUrl", None)
    config["cspIngressHostName"] = os.getenv("cspIngressHostName", None)
    config["appsNamespace"] = os.getenv("appsNamespace", None)
    config["infraNamespace"] = os.getenv("infraNamespace", None)
    config["createServiceOrg"] = False

    # Service org
    config["orgId"] = os.getenv("orgId", None)
    config["serviceOrgDisplayName"] = os.getenv("serviceOrgDisplayName", None)

    # Service definition
    config["env"] = os.getenv("env", None)
    config["serviceName"] = os.getenv("serviceName", None)
    config["serviceDisplayName"] = os.getenv("serviceDisplayName", None)
    config["instanceDisplayName"] = os.getenv("instanceDisplayName", None)
    config["serviceUrl"] = os.getenv("serviceUrl", None)
    config["serviceTicker"] = os.getenv("serviceTicker", None)
    config["productIdentifier"] = os.getenv("productIdentifier", None)
    config["serviceDescription"] = os.getenv("serviceDescription", None)
    config["isBeta"] = os.getenv("isBeta", "false")
    config["isPrimary"] = os.getenv("isPrimary", "false")
    config["isGated"] = os.getenv("isGated", "true")
    config["isCheckoutEnabled"] = os.getenv("isCheckoutEnabled", "false")
    config["serviceType"] = os.getenv("serviceType", None)
    config["defaultBillingEngine"] = os.getenv("defaultBillingEngine", None)
    config["isVisible"] = os.getenv("isVisible", "true")

    # Service instance
    config["serviceInstanceId"] = os.getenv("serviceInstanceId", None)
    config["serviceInstanceIdCSR"] = os.getenv("serviceInstanceIdCSR", None)

    # Confer org
    config["createConferOrg"] = False
    config["conferOrgId"] = os.getenv("conferOrgId", None)
    config["conferUrl"] = os.getenv("conferUrl", None)
    config["conferDisplayName"] = os.getenv("conferDisplayName", None)

    # Customer org
    config["customerUrl"] = os.getenv("customerUrl")
    config["createCustomerOrg"] = os.getenv("createCustomerOrg", "False") in ["True", "true", "yes", "Yes"]
    config["customerOrgDisplayName"] = os.getenv("customerOrgDisplayName", None)

    # Extra
    config["useCSPCredentials"] = False

    return verify_config()


def verify_config():
    required_keys = set()

    if config["destroyMode"]:
        # Will most likely require cspUser and cspPassword as well
        required_keys.update({ "cspBootstrapAPIToken", "cspUrl", "appsNamespace", "infraNamespace" })
        return check_required_keys(required_keys)

    if not config["orgId"]:
        # When creating a service org, username and password have to be provided, in order to generate an access token for the new org.
        required_keys.update({ "serviceOrgDisplayName", "cspUser", "cspPassword" })
        logging.info("'orgId' is not set, a service org will be created")
        config["createServiceOrg"] = True

    required_keys.update({ "cspBootstrapAPIToken",
                           "cspUrl",
                           "cspIngressHostName",
                           "appsNamespace",
                           "infraNamespace",
                           "manifestType",
                           "env",
                           "serviceName",
                           "serviceDisplayName",
                           "instanceDisplayName",
                           "serviceUrl",
                           "serviceDescription",
                           "customerUrl" })

    if not config["conferOrgId"]:
        logging.info("'conferOrgId' has not been set. A CSR organization will be created.")
        config["createConferOrg"] = True
        required_keys.update({ "conferUrl", "conferDisplayName", "cspUser", "cspPassword" })

    if not config["cspBootstrapAPITokenCSR"]:
        # If an API token for the confer org is not provided, a user and password are needed, in order to finish the bootstrapping
        required_keys.update({ "cspUser", "cspPassword" })

    if config["createCustomerOrg"]:
        logging.info("'createCustomerOrg' has been set to true. A customer organization will be created.")
        required_keys.update({ "customerUrl", "customerOrgDisplayName" })


    if "cspUser" in required_keys and "cspPassword" in required_keys:
        config["useCSPCredentials"] = True

    return check_required_keys(required_keys)


def check_required_keys(required_keys):
    for key in required_keys:
        if not config[key]:
            raise Exception("Required variable '{}' is not set.".format(key))
    return True


def populate_config_map(org_id, service_definition_id,
                        service_instance_id, service_instance_id_csr,
                        service_ticker, confer_org_id,
                        support_feature_group_id,
                        support_role_name, customer_org_id, clm_notification_message_type,
                        #clm_csp_notification_publisher_id,
                        identity_csp_notification_subscriber_id):

    # Defines the output ConfigMap ('csp-bootstrap-outputs')
    config_map = {
        "service_org_id": org_id,
        "service_definition_id": service_definition_id,
        "service_instance_id": service_instance_id,
        "service_instance_id_csr": service_instance_id_csr,
        "web_app_client_secret_name": "cspoauth",
        "service_client_secret_name": "server-csp-client",
        "clm_csp_service_owner_connector_secret_name": "clm-csp-service-owner-connector",
        "clm_csp_operator_org_connector_secret_name": "clm-csp-operator-org-connector",
        "confer_org_id": confer_org_id,
        "support_feature_group_id": support_feature_group_id,
        "support_feature_role": support_role_name,
        "clm_notification_message_type": clm_notification_message_type,
        #"clm_csp_notification_publisher_id": clm_csp_notification_publisher_id,
        "identity_csp_notification_subscriber_id": identity_csp_notification_subscriber_id
    }

    if service_ticker:
        config_map.update({ "service_ticker": service_ticker })
    if customer_org_id:
        config_map.update({ "customer_org_id": customer_org_id })

    return config_map


def cleanup_all(csp_api, kubernetes_client, outputs_config_map_name):
    # Might need to create a temp OAuth app here too, in order to delete stuff

    try:
        client_id_web_app = b64decode(kubernetes_client.get_secret("cspoauth", config["appsNamespace"])['client_id']).decode('utf-8')
        kubernetes_client.delete_secret("cspoauth", config["appsNamespace"])
        kubernetes_client.delete_secret("cspoauth", config["infraNamespace"])
        csp_api.delete_oauth_apps(config["orgId"], [client_id_web_app])
        client_id_server_app = b64decode(kubernetes_client.get_secret("server-csp-client", config["appsNamespace"])['client_id']).decode('utf-8')
        kubernetes_client.delete_secret("server-csp-client", config["appsNamespace"])
        kubernetes_client.delete_secret("server-csp-client", config["infraNamespace"])
        csp_api.delete_oauth_apps(config["orgId"], [client_id_server_app])
    except Exception as e:
        print(e)
        logging.info("Skipping OAuth apps resources clean-up")

    kubernetes_client.delete_config_map(outputs_config_map_name, config["appsNamespace"])
    kubernetes_client.delete_config_map(outputs_config_map_name, config["infraNamespace"])
    logging.info("Done")


def main():
    populate_config()

    try:
        kubernetes_client = KubernetesClient()
    except Exception as e:
        logging.critical("Failed to create Kubernetes Client")
        raise e

    csp = CSP(config["cspUrl"], grant_type="api_token",
              api_token=config["cspBootstrapAPIToken"], kubernetes_client=kubernetes_client)

    outputs_cm_name = "csp-bootstrap-outputs"

    if config["destroyMode"]:
        cleanup_all(csp, kubernetes_client, outputs_cm_name)
        return 0

    if config["createServiceOrg"]:
        try:
            config["orgId"] = csp.create_org(config["serviceOrgDisplayName"], create_service_org=True)
        except:
            raise Exception("'orgId' could not be acquired")

    if config["createConferOrg"]:
        try:
            config["conferOrgId"] = csp.create_org(config["conferDisplayName"])
        except:
            raise Exception("'conferOrgId' could not be acquired")

    # Create temp OAuth app if user and password have to be used
    if config["useCSPCredentials"]:
        current_user_info = csp.get_user_info()
        temp_oauth_app = csp.create_temp_oauth_app(config["env"], current_user_info["context_name"])
        csp.set_auth(grant_type="password", client_id=temp_oauth_app["client_id"],
                     client_secret=temp_oauth_app["client_secret"], username=config["cspUser"], password=config["cspPassword"], orgId=config["orgId"])
        # Register atexit function to delete the temp OAuth app
        atexit.register(csp.delete_oauth_apps, current_user_info["context_name"], [ temp_oauth_app["client_id"] ])

    # Setup one more client for CSR org access
    if config["cspBootstrapAPITokenCSR"]:
        csp_csr = CSP(config["cspUrl"], grant_type="api_token",
                      api_token=config["cspBootstrapAPITokenCSR"],
                      kubernetes_client=kubernetes_client)
    else:
        csp_csr = CSP(config["cspUrl"], grant_type="password", client_id=temp_oauth_app["client_id"],
                      client_secret=temp_oauth_app["client_secret"], username=config["cspUser"],
                      password=config["cspPassword"], orgId=config["conferOrgId"],
                      kubernetes_client=kubernetes_client)

    if config["createServiceOrg"]:
        csp.add_service_client_to_org(config["orgId"], temp_oauth_app["client_id"], org_roles=["org_owner"],
                                      service_roles=[])


    # Try to get service_definition_id from either outputs ConfigMap, CSP or create it
    service_definition_id = kubernetes_client.get_key_from_config_map("service_definition_id",
                                                                      outputs_cm_name,
                                                                      config["appsNamespace"])

    definition_template_file = "resource/service_definition_template.yaml.j2"
    extra_entities_template_file = "resource/roles_with_permissions.yaml.j2"
    service_icon_file = "resource/vmw-ico-carbonblackcloud-rgb_16px.svg"
    request_access_url = "https://www.vmware.com/resources/security/demo.html"
    pricing_info_url = "https://www.vmware.com/products/carbon-black-cloud.html"
    if not service_definition_id:
        service_definition_id = csp.get_service_definition_id(config["serviceName"])

    if not service_definition_id:
        definition_values = { "serviceName": config["serviceName"],
                              "serviceDisplayName": config["serviceDisplayName"],
                              "serviceTicker": config["serviceTicker"],
                              "productIdentifier": config["productIdentifier"],
                              "serviceDescription": config["serviceDescription"],
                              "orgId": config["orgId"],
                              "isBeta": config["isBeta"],
                              "isPrimary": config["isPrimary"],
                              "isGated": config["isGated"],
                              "isCheckoutEnabled": config["isCheckoutEnabled"],
                              "serviceType": config["serviceType"],
                              "serviceUrl": config["serviceUrl"],
                              "defaultBillingEngine": config["defaultBillingEngine"],
                              "isVisible": config["isVisible"] }
        service_definition_id = csp.create_service_definition(definition_template_file, definition_values)
    else:
        logging.info("Definition ID for service '{}' already exists.".format(config["serviceName"]))

    # Patch service roles if there are changes in the template
    csp.update_service_definition_roles(service_definition_id, definition_template_file,
                                        extra_entities_template_file,
                                        config["manifestType"],
                                        csp.get_service_ticker(config["serviceName"]))
    # Set service icon and patch service URLs
    csp.set_service_icon(service_definition_id, service_icon_file)
    csp.patch_service_urls(service_definition_id, request_access_url, pricing_info_url)

    # Creates or replicates all OAuth apps-related secrets in both apps/infra namespaces
    oauth_apps = {  "cspoauth": { "exists_in": [] },
                    "server-csp-client": { "exists_in": [] },
                    "clm-csp-service-owner-connector": { "exists_in": [] },
                    "clm-csp-operator-org-connector": { "exists_in": [] },
                    "common-operator-org-connector": { "exists_in": [] }
    }

    for ns in [ config["appsNamespace"], config["infraNamespace"] ]:
        for app in oauth_apps.keys():
            k8s_secret = kubernetes_client.get_secret(app, ns)
            if k8s_secret:
                oauth_apps[app]["exists_in"].append(ns)
                if "secret" not in oauth_apps[app]:
                    logging.info("Got secret for '{}' from ns '{}'".format(app, ns))
                    oauth_apps[app]["secret"] = { "client_id": b64decode(k8s_secret["client_id"]).decode("utf-8"), "client_secret": b64decode(k8s_secret["client_secret"]).decode("utf-8") }

    for app in oauth_apps.keys():
        # Delete any k8s secrets, whose OAuth apps can't be found in the Service Org
        if "secret" in oauth_apps[app]:
            app_org_id = config["orgId"]
            # The CLM operator org OAuth app is registered in the CSR org as well as the "common connector".
            if app == "clm-csp-operator-org-connector" or app == "common-operator-org-connector":
                app_org_id = config["conferOrgId"]
            if not csp.get_oauth_app(app_org_id, oauth_apps[app]["secret"]["client_id"]):
                for ns in [ config["appsNamespace"], config["infraNamespace"] ]:
                    kubernetes_client.delete_secret(app, ns)
                oauth_apps[app].pop("secret")

        # Create any OAuth apps that are missing
        if "secret" not in oauth_apps[app]:
            oauth_apps[app]["exists_in"] += [ config["appsNamespace"], config["infraNamespace"] ]
            sct = {}
            if app == "cspoauth":
                sct = csp.create_web_app_client(config["env"],
                                                config["orgId"],
                                                config["conferUrl"],
                                                config["customerUrl"],
                                                service_definition_id,
                                                config["appsNamespace"],
                                                config["infraNamespace"])
            if app == "server-csp-client":
                sct = csp.create_service_client(config["env"],
                                                config["orgId"],
                                                service_definition_id,
                                                config["appsNamespace"],
                                                config["infraNamespace"])
            if app == "clm-csp-service-owner-connector":
                sct = csp.create_oauth_app(app,
                                           "{} CLM service owner connector".format(config["env"]),
                                           "CLM service owner connector",
                                           config["orgId"],
                                           service_definition_id,
                                           config["appsNamespace"],
                                           config["infraNamespace"])
            if app == "clm-csp-operator-org-connector":
                sct = csp_csr.create_oauth_app(app,
                                               "{} CLM operator org connector".format(config["env"]),
                                               "CLM operator org connector",
                                               config["conferOrgId"],
                                               service_definition_id,
                                               config["appsNamespace"],
                                               config["infraNamespace"])
            if app == "common-operator-org-connector":
                sct = csp_csr.create_oauth_app(app,
                                               "{} common operator org connector".format(config["env"]),
                                               "Common operator org connector",
                                               config["conferOrgId"],
                                               service_definition_id,
                                               config["appsNamespace"],
                                               config["infraNamespace"])
            oauth_apps[app]["secret"] = { "client_id": sct[0], "client_secret": sct[1] }

        # Replicate the missing secrets
        for ns in [ config["appsNamespace"], config["infraNamespace"] ]:
            if ns not in oauth_apps[app]["exists_in"]:
                kubernetes_client.create_secret(app, oauth_apps[app]["secret"], ns)

    # Service instances creation
    # Normal
    old_format = "{} - {}".format(config["serviceDisplayName"], config["env"])
    old_format2 = config["env"]
    new_format = config["instanceDisplayName"]
    if not config["serviceInstanceId"]:
        config["serviceInstanceId"] = csp.get_service_instance_id(service_definition_id, old_format)
        if not config["serviceInstanceId"]:
            config["serviceInstanceId"] = csp.get_service_instance_id(service_definition_id, old_format2)
        if not config["serviceInstanceId"]:
            config["serviceInstanceId"] = csp.get_service_instance_id(service_definition_id, new_format)
        if not config["serviceInstanceId"]:
            config["serviceInstanceId"] = csp.create_service_instance(service_definition_id,
                                                                      new_format,
                                                                      config["serviceUrl"])

    current_name = csp.get_service_instance(service_definition_id,
                                            config["serviceInstanceId"])["displayName"]
    if current_name != new_format:
        csp.patch_service_instance(service_definition_id,
                                   config["serviceInstanceId"],
                                   new_format)

    # CSR
    old_format_csr = "{} - {} - CSR".format(config["serviceDisplayName"], config["env"])
    old_format_csr2 = "{} - CSR".format(config["env"])
    new_format_csr = "{} - CSR".format(config["instanceDisplayName"])
    if not config["serviceInstanceIdCSR"]:
        config["serviceInstanceIdCSR"] = csp.get_service_instance_id(service_definition_id, old_format_csr)
        if not config["serviceInstanceIdCSR"]:
            config["serviceInstanceIdCSR"] = csp.get_service_instance_id(service_definition_id, old_format_csr2)
        if not config["serviceInstanceIdCSR"]:
            config["serviceInstanceIdCSR"] = csp.get_service_instance_id(service_definition_id, new_format_csr)
        if not config["serviceInstanceIdCSR"]:
            config["serviceInstanceIdCSR"] = csp.create_service_instance(service_definition_id,
                                                                         new_format_csr,
                                                                         config["conferUrl"])

    current_name_csr = csp.get_service_instance(service_definition_id,
                                                config["serviceInstanceIdCSR"])["displayName"]
    if current_name_csr != new_format_csr:
        csp.patch_service_instance(service_definition_id,
                                   config["serviceInstanceIdCSR"],
                                   new_format_csr)

    # Granting access
    csp.grant_service_access_to_org(service_definition_id, config["serviceInstanceId"], config["orgId"])
    csp.grant_service_access_to_org(service_definition_id, config["serviceInstanceIdCSR"], config["conferOrgId"])

    csp.assign_org_to_service_instance(service_definition_id, config["conferOrgId"],
                                       config["serviceInstanceIdCSR"])
    csp.assign_org_to_service_instance(service_definition_id, config["orgId"],
                                       config["serviceInstanceId"])
    csp.add_service_client_to_org(config["orgId"],
                                  oauth_apps["server-csp-client"]["secret"]["client_id"],
                                  org_roles=["org_owner", "service_owner", "service_enabler"],
                                  service_roles=["psc.role:BETA_SUPER_ADMIN"],
                                  service_definition_id=service_definition_id)
    csp.add_service_client_to_org(config["orgId"],
                                  oauth_apps["clm-csp-service-owner-connector"]["secret"]["client_id"],
                                  org_roles=["org_member", "service_owner"],
                                  service_roles=["psc.role:BETA_SUPER_ADMIN"],
                                  service_definition_id=service_definition_id)

    # Make the CSR org an operator for the new service
    csp.make_org_operator(service_definition_id, config["conferOrgId"])

    # Create customer org (used in dev scenarios only)
    customer_org_id = ""
    if config["createCustomerOrg"]:
        customer_org_id = csp.get_org_id(config["customerOrgDisplayName"])
        if not customer_org_id:
            customer_org_id = csp.create_org(config["customerOrgDisplayName"])
        customer_service_invitation_id = csp.create_service_invitation(service_definition_id,
                                                                       config["serviceInstanceId"])
        csp.redeem_service_invitation(customer_service_invitation_id, customer_org_id)

    # Operator org operations
    support_role_name = "psc.role:SUPPORT_FEATURE"
    # Add OAuth apps to CSR organization
    csp_csr.add_service_client_to_org(config["conferOrgId"],
                                      oauth_apps["server-csp-client"]["secret"]["client_id"],
                                      org_roles=["org_owner"], service_roles=[support_role_name],
                                      service_definition_id=service_definition_id)
    csp_csr.add_service_client_to_org(config["conferOrgId"],
                                      oauth_apps["clm-csp-operator-org-connector"]["secret"]["client_id"],
                                      org_roles=["org_member"],
                                      service_roles=["psc.role:PROVISIONING", 
                                                     "psc.role:ALL_AND_CSR"],
                                      service_definition_id=service_definition_id)
    csp_csr.add_service_client_to_org(config["conferOrgId"],
                                      oauth_apps["common-operator-org-connector"]["secret"]["client_id"],
                                      org_roles=["org_owner"],
                                      service_roles=[ r["name"] for r in csp.get_service_roles(service_definition_id) ],
                                      service_definition_id=service_definition_id)


    # Create Support Group and add the roles
    support_group_name = "Support Group"
    support_feature_group_id = csp.get_group(config["conferOrgId"], name=support_group_name)
    if not support_feature_group_id:
        support_feature_group_id = csp.create_group(org_id=config["conferOrgId"],
                                                    name=support_group_name,
                                                    description="Group for support users")
    csp_csr.update_group_roles(config["conferOrgId"], service_definition_id,
                               config["serviceInstanceIdCSR"], support_feature_group_id,
                               org_roles_to_add=["org_member"],
                               # no need for svc roles
                               service_roles_to_add=["psc.role:BETA_VIEW_ALL"])

    ### CLM and Notifications steps
    clm_notification_message_type = "cbc-customer-lifecycle-event"
    #clm_message_type_attributes_file = "resource/clm_message_type_attributes.json"
    #clm_csp_notification_publisher_id = kubernetes_client.get_key_from_config_map("clm_csp_notification_publisher_id",
    #                                                                              outputs_cm_name,
    #                                                                              config["appsNamespace"])

    #clm_svc_owner_client_id = oauth_apps["clm-csp-service-owner-connector"]["secret"]["client_id"]
    #create_publisher = False
    #if not clm_csp_notification_publisher_id:
    #    clm_csp_notification_publisher = csp.get_publisher(config["orgId"],
    #                                                       clm_notification_message_type)
    #    if not clm_csp_notification_publisher:
    #        create_publisher = True
    #    elif clm_csp_notification_publisher["clientID"] != clm_svc_owner_client_id:
    #        diff_pub_msg = "Publisher '{}' has client id '{}', which is different than '{}'."
    #        logging.info(diff_pub_msg.format(clm_csp_notification_publisher["publisherID"],
    #                                         clm_csp_notification_publisher["clientID"],
    #                                         clm_svc_owner_client_id))
    #        csp.delete_publisher(clm_csp_notification_publisher["publisherID"])
    #        create_publisher = True

    #if create_publisher:
    #    clm_csp_notification_publisher_id = csp.register_publisher(config["orgId"],
    #                                                               clm_svc_owner_client_id,
    #                                                               [ clm_notification_message_type ])

    #if not csp.get_message_type(clm_notification_message_type, clm_csp_notification_publisher_id):
    #    csp.register_message_type(clm_notification_message_type, clm_csp_notification_publisher_id)

    #csp.create_message_type_attributes(clm_notification_message_type, clm_message_type_attributes_file,
    #                                       clm_csp_notification_publisher_id)
    #csp.update_message_type_attributes(clm_notification_message_type,
    #                                   clm_message_type_attributes_file,
    #                                   clm_csp_notification_publisher_id)

    # Re-create subscriber on every bootstrap run
    identity_csp_notification_subscriber_id = csp.get_subscriber(oauth_apps["server-csp-client"]["secret"]["client_id"],
                                                                 config["orgId"],
                                                                 clm_notification_message_type)

    if identity_csp_notification_subscriber_id:
        csp.delete_subscriber(identity_csp_notification_subscriber_id)

    identity_csp_notification_subscriber_id = csp.register_subscriber(oauth_apps["server-csp-client"]["secret"]["client_id"],
                                                                      config["orgId"])

    csp.update_subscriber_auth(identity_csp_notification_subscriber_id,
                               oauth_apps["server-csp-client"]["secret"])
    identity_notifications_endpoint = "{}/identity/v1/notifications/csp/".format(config["cspIngressHostName"])
    # Create all identity message type subscriptions
    identity_subscriptions_file = "resource/identity_subscription_message_types.yaml.j2"
    csp.create_identity_subscriptions(identity_subscriptions_file,
                                      identity_csp_notification_subscriber_id,
                                      oauth_apps["server-csp-client"]["secret"]["client_id"],
                                      identity_notifications_endpoint,
                                      #config["env"],
                                      config["manifestType"])

    config_map = populate_config_map(config["orgId"],
                                     service_definition_id,
                                     config["serviceInstanceId"],
                                     config["serviceInstanceIdCSR"],
                                     csp.get_service_ticker(config["serviceName"]),
                                     config["conferOrgId"],
                                     support_feature_group_id,
                                     support_role_name,
                                     customer_org_id,
                                     clm_notification_message_type,
                                     #clm_csp_notification_publisher_id,
                                     identity_csp_notification_subscriber_id)


    config_map_apps = kubernetes_client.get_config_map(outputs_cm_name, config["appsNamespace"])
    config_map_infra = kubernetes_client.get_config_map(outputs_cm_name, config["infraNamespace"])

    if config_map_apps:
        kubernetes_client.patch_config_map(outputs_cm_name, config_map, config["appsNamespace"])
    else:
        kubernetes_client.create_config_map(outputs_cm_name, config_map, config["appsNamespace"])

    if config_map_infra:
        kubernetes_client.patch_config_map(outputs_cm_name, config_map, config["infraNamespace"])
    else:
        kubernetes_client.create_config_map(outputs_cm_name, config_map, config["infraNamespace"])

    return 0


if __name__ == "__main__":
    main()
