import json
import logging
import random
import string
import yaml

import jinja2

from kube_client import *
from sessions import *


class CSP:
    def __init__(self, csp_url, kubernetes_client=None, **kwargs):
        self.session = CSPSession(base_url=csp_url, **kwargs)
        self.kubernetes_client = kubernetes_client


    def get_service_definition_id(self, service_name):
        path = "/csp/gateway/slc/api/definitions"
        params = { "expand": "true" }

        r = self.session.get(path, allow_redirects=True, params=params)
        resp = json.loads(r.text)
        for service in resp['results']:
            if service['name'] == service_name:
                return service['documentSelfLink'].split('/')[-1]
        return None


    def set_auth(self, **kwargs):
        self.session.authenticate(**kwargs)


    def __get_user_orgs(self, user_id):
        path = "/csp/gateway/am/api/v2/loggedin/user/orgs"
        headers = self.session.headers.copy()
        headers.update({ "csp-auth-token": self.session.headers["Authorization"].split(' ')[-1] })

        logging.info("Get user '{}' orgs".format(user_id))
        r = self.session.get(path, headers=headers, allow_redirects=True)
        return json.loads(r.text)["results"]


    def get_service_ticker(self, service_name):
        path = "/csp/gateway/slc/api/definitions"
        params = { "expand": "true" }

        r = self.session.get(path, allow_redirects=True, params=params)
        resp = json.loads(r.text)
        for service in resp['results']:
            if service['name'] == service_name:
                return service['serviceTicker']
        return None


    def __service_ticker_exists(self, ticker):
        path = "/csp/gateway/slc/api/definitions"
        params = { "expand": "true" }

        r = self.session.get(path, allow_redirects=True, params=params)
        resp = json.loads(r.text)
        for service in resp['results']:
            if service['serviceTicker'] == ticker:
                return True
        return False


    def __generate_service_ticker(self):
        while True:
            ticker_len = random.randint(2, 5)
            service_ticker = ''.join(random.choices(string.ascii_lowercase, k=ticker_len))
            if not self.__service_ticker_exists(service_ticker) and service_ticker not in ["csp", "gaz", "custom"]:
                return service_ticker


    def __generate_service_definition_yaml(self, template_file, definition_values):
        # Generate a service ticker if one was not provided
        if not definition_values['serviceTicker']:
            logging.info("Generating service ticker, because none was provided")
            definition_values.update({ "serviceTicker": self.__generate_service_ticker() })

        templateLoader = jinja2.FileSystemLoader(searchpath="./")
        templateEnv = jinja2.Environment(loader=templateLoader)
        template = templateEnv.get_template(template_file)

        outputYaml = template.render(definition_values)
        return yaml.safe_load(outputYaml)


    def create_service_definition(self, template_file, definition_values):
        path = "/csp/gateway/slc/api/definitions"
        headers = self.session.headers.copy()
        headers.update({ "Content-Type": "text/plain" })

        definition_yaml = self.__generate_service_definition_yaml(template_file, definition_values)
        logging.info("Creating service definition")
        r = self.session.post(path, allow_redirects=True, headers=headers, data=yaml.dump(definition_yaml))
        logging.info(r.text)
        resp = json.loads(r.text)
        # Returns the "service_definition_id"
        return resp['refLink'].split('/')[-1]


    def get_user_info(self):
        path = "/csp/gateway/am/api/userinfo"
        logging.info("Getting current user info")
        r = self.session.get(path, allow_redirects=True)
        if r.ok:
            return json.loads(r.text)
        else:
            return None


    def grant_service_access_to_org(self, service_definition_id, service_instance_id, org_id):
        path = "/csp/gateway/slc/api/service-access"
        payload = { "serviceDefinitionId": service_definition_id,
                    "orgId": org_id,
                    "isTosPreSigned": "true",
                    "instanceIds": [ service_instance_id ] }

        logging.info("Granting access for definition '{}' and instance '{}' to organization '{}'".format(service_definition_id, service_instance_id, org_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def delete_service_definition(self, service_definition_id):
        path = "/csp/gateway/slc/api/definitions/external/{}".format(service_definition_id)
        logging.info("Deleting service definition with id '{}'".format(service_definition_id))
        r = self.session.delete(path, allow_redirects=True)
        logging.info(r.text)


    def delete_oauth_apps(self, org_id, client_ids):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps".format(org_id)
        payload = { "clientIdsToDelete": client_ids }

        logging.info("Deleting OAuth apps {}".format(client_ids))
        r = self.session.delete(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def create_web_app_client(self, env, org_id, confer_url, customer_url, service_definition_id, apps_namespace, infra_namespace):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps".format(org_id)
        payload = {
              "displayName": env + "-integration",
              "organizationId": org_id,
              "maxGroupsInIdToken": 20,
              "redirectUris": [
                  confer_url + "/auth/v2/csp/callback",
                  customer_url + "/auth/v2/csp/callback"
              ],
              "forcePkce": "false",
              "allowOpenRedirectUris": "false",
              "grantTypes": [
                  "refresh_token",
                  "authorization_code"
              ],
              "accessTokenTTL": 900,
              "description": env + " Environment.",
              "allowedScopes": {
                  "generalScopes": [
                      "customer_number",
                      "openid",
                      "group_ids",
                      "group_names"
                  ],
                  "organizationScopes": {
                      "keptInToken": [
                          "ROLES"
                      ],
                      "allRoles": "false",
                      "roleNames": [
                          "service_enabler",
                          "service_owner",
                          "org_owner"
                      ],
                      "permissions": [],
                      "allPermissions": "false"
                  },
                  "servicesScopes": [
                              {
                                  "allPermissions": "true",
                                  "keptInToken": [
                                      "ROLES",
                                      "PERMISSIONS"
                                  ],
                                  "serviceDefinitionId": service_definition_id,
                                  "allRoles": "true"
                              }
                          ],
                  "allRoles": "false"
              },
              "publicClient": "false",
              "refreshTokenTTL": 15552000
        }

        logging.info("Creating Web App client")
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if self.kubernetes_client:
            secret_data = { "client_id": resp['clientId'], "client_secret": resp['clientSecret'] }
            self.kubernetes_client.create_secret("cspoauth", secret_data, apps_namespace)
            self.kubernetes_client.create_secret("cspoauth", secret_data, infra_namespace)
        return resp['clientId'], resp['clientSecret']


    def create_service_client(self, env, org_id, service_definition_id, apps_namespace, infra_namespace):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps".format(org_id)
        payload = {
           "displayName": env + " service client",
           "description": "App used by CBC for server to server communication",
           "grantTypes": [
               "client_credentials"
           ],
           "accessTokenTTL": 1800,
           "refreshTokenTTL": 7776000,
           "allowedScopes": {
               "allRoles": "false",
               "organizationScopes": {
                   "allRoles": "true",
                   "roleNames": [],
                   "permissions": [],
                   "allPermissions": "true",
                   "keptInToken": [
                       "ROLES"
                   ]
               },
               "servicesScopes": [
                    {
                        "allPermissions": "false",
                        "keptInToken": [
                            "ROLES"
                        ],
                        "serviceDefinitionId": service_definition_id,
                        "allRoles": "true"
                    }
                ],
               "generalScopes": []
           },
           "organizationId": org_id,
           "maxGroupsInIdToken": 20
        }

        logging.info("Creating service client")
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if self.kubernetes_client:
            secret_data = { "client_id": resp['clientId'], "client_secret": resp['clientSecret'] }
            self.kubernetes_client.create_secret("server-csp-client", secret_data, apps_namespace)
            self.kubernetes_client.create_secret("server-csp-client", secret_data, infra_namespace)
        return resp['clientId'], resp['clientSecret']


    def create_oauth_app(self, name, display_name, description, org_id, service_definition_id, apps_namespace, infra_namespace):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps".format(org_id)
        payload = {
           "displayName": display_name,
           "description": description,
           "grantTypes": [
               "client_credentials"
           ],
           "accessTokenTTL": 1800,
           "refreshTokenTTL": 7776000,
           "allowedScopes": {
               "allRoles": "false",
               "organizationScopes": {
                   "allRoles": "true",
                   "roleNames": [],
                   "permissions": [],
                   "allPermissions": "true",
                   "keptInToken": [
                       "ROLES"
                   ]
               },
               "servicesScopes": [
                    {
                        "allPermissions": "true",
                        "keptInToken": [
                            "ROLES"
                        ],
                        "serviceDefinitionId": service_definition_id,
                        "allRoles": "true"
                    }
                ],
               "generalScopes": []
           },
           "organizationId": org_id,
           "maxGroupsInIdToken": 20
        }

        logging.info("Creating OAuth app '{}'".format(name))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if self.kubernetes_client:
            secret_data = { "client_id": resp['clientId'], "client_secret": resp['clientSecret'] }
            self.kubernetes_client.create_secret(name, secret_data, apps_namespace)
            self.kubernetes_client.create_secret(name, secret_data, infra_namespace)
        return resp['clientId'], resp['clientSecret']


    def get_oauth_app(self, org_id, client_id):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps/{}".format(org_id, client_id)
        logging.info("Check if OAuth app with client id '{}' exists.".format(client_id))
        r = self.session.get(path, allow_redirects=True)
        res = None
        if r.status_code == 200:
            logging.info("OAuth app with client id '{}' exists.".format(client_id))
            res = json.loads(r.text)
        elif r.status_code == 404:
            logging.info("OAuth app with client id '{}' not found".format(client_id))
        else:
            logging.error(r.text)
            raise RuntimeError()

        return res


    def register_publisher(self, org_id, client_id, message_types):
        path = "/csp/gateway/po/api/v1/publishers"
        payload = {
                      "clientID": client_id,
                      "messageTypes": message_types,
                      "orgID": org_id,
                      "role": "org_owner"
        }

        logging.info("Registering publisher for message types \"{}\" for client '{}' in org '{}'".format(message_types, client_id, org_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if r.status_code != 201:
            logging.error("Error creating publisher.")
            raise RuntimeError()
        else:
            return resp["publisherID"]


    def get_publisher(self, org_id, message_type):
        path = "/csp/gateway/po/api/v1/publishers"
        parameters = { "orgID": org_id, "messageType": message_type }
        logging.info("Get publisher for message type '{}' in org '{}'".format(message_type, org_id))
        r = self.session.get(path, allow_redirects=True, params=parameters)
        resp = json.loads(r.text)
        if "messageTypes" in resp and message_type in resp["messageTypes"]:
            logging.info("Publisher found - '{}'".format(resp["publisherID"]))
            return resp
        else:
            logging.info("Pubisher not found")
            return None


    def delete_publisher(self, publisher_id):
        path = "/csp/gateway/po/api/v1/publishers/{}".format(publisher_id)
        logging.info("Delete publisher '{}'".format(publisher_id))
        r = self.session.delete(path, allow_redirects=True)
        logging.info(r.text)


    def register_message_type(self, name, publisher_id):
        path = "/csp/gateway/po/api/v1/message-types"
        payload = {
                      "description": "CBC CLM Notification of customer lifecycle event",
                      "displayName": "CBC Customer Lifecycle Event",
                      "messageType": name,
                      "publisherID": publisher_id
        }
        logging.info("Creating CSP notifications message type '{}'".format(name))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if r.status_code != 201:
            logging.info("Error creating message type.")
            raise RuntimeError()


    def get_message_type(self, name, publisher_id):
        path = "/csp/gateway/po/api/v1/message-types"
        parameters = { "publisherID": publisher_id, "messageType": name }
        logging.info("Get message type '{}' for publisher '{}'".format(name, publisher_id))
        r = self.session.get(path, allow_redirects=True, params=parameters)
        resp = json.loads(r.text)
        if r.status_code == 200:
            if resp:
                logging.info("Message type '{}' already exists".format(name))
                return resp
            else:
                logging.info("Message tpye '{}' does not exist".format(name))
                return None
        else:
            logging.info("Error getting the correct message type")
            raise RuntimeError(resp)


    def update_message_type_attributes(self, message_type, message_type_filename, publisher_id):
        path = "/csp/gateway/po/api/v1/message-types/{}/attributes".format(message_type)
        exst_resp = self.session.get(path, allow_redirects=True)
        try:
            existing = json.loads(exst_resp.text)
        except:
            existing = dict()

        message_type_file = open(message_type_filename)
        template_attributes = json.load(message_type_file)["attributes"]
        logging.info("Patching attributes for message type '{}'".format(message_type))
        to_delete = [ ak for ak in existing if ak not in [ ta for ta in template_attributes ] ]
        to_create = { ak:av for ak, av in template_attributes.items() if ak not in [ ea for ea in existing.keys() ] }
        logging.info("Atrributes to delete: {}".format(to_delete))
        logging.info("Attributes to add: {}".format(to_create))

        payload = {
                    "attributes": {
                        "attributesToAdd": to_create,
                        "attributesToRemove": to_delete
                    },
                    "publisherID": publisher_id
        }

        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        resp = json.loads(r.text)
        if r.status_code != 202:
            logging.info("Error patching attributes for message type '{}'".format(message_type))
            raise RuntimeError(resp)
        else:
            logging.info("Status code '{}'".format(r.status_code))


    def create_message_type_attributes(self, message_type, message_type_filename, publisher_id):
        path = "/csp/gateway/po/api/v1/message-types/{}/attributes".format(message_type)
        message_type_file = open(message_type_filename)
        payload = json.load(message_type_file)
        payload["publisherID"] = publisher_id

        logging.info("Creating attributes for message type '{}'".format(message_type))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(json.loads(r.text))
        logging.info("Status code '{}'".format(r.status_code))


    def register_subscriber(self, client_id, org_id):
        path = "/csp/gateway/po/api/v1/subscribers"
        payload = {
                      "clientID": client_id,
                      "orgID": org_id,
                      "role": "org_owner"
        }

        logging.info("Registering subscriber with client id '{}' for org '{}'".format(client_id, org_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        if r.status_code != 201:
            logging.error("Error creating subscriber.")
            raise RuntimeError()
        else:
            return resp["subscriberID"]


    def get_subscriber(self, client_id, org_id, message_type):
        path = "/csp/gateway/po/api/v1/subscribers"
        parameters = { "messageType": message_type, "clientID": client_id }
        logging.info("Get subscriber for message type '{}' for client '{}' in org '{}'".format(message_type, client_id, org_id))
        r = self.session.get(path, allow_redirects=True, params=parameters)
        resp = json.loads(r.text)
        if r.status_code == 200:
            for s in resp:
                if s["orgID"] == org_id:
                    logging.info("Subscriber found")
                    return s["subscriberID"]
        elif r.status_code == 404:
            logging.info("Subscriber not found")
        else:
            logging.error(resp)

        return None


    def delete_subscriber(self, subscriber_id):
        path = "/csp/gateway/po/api/v1/subscribers/{}".format(subscriber_id)
        logging.info("Deleting subscriber '{}'".format(subscriber_id))
        r = self.session.delete(path, allow_redirects=True)
        logging.info("Status code '{}'".format(r.status_code))


    def update_subscriber_auth(self, subscriber_id, auth):
        path = "/csp/gateway/po/api/v1/subscribers/{}/auth".format(subscriber_id)
        payload = {
                      "authType": "CSP-CLIENT",
                      "authParams": {
                        "authTokenName": "Authorization",
                        "clientId": auth["client_id"],
                        "clientSecret": auth["client_secret"],
                        "tokenPrefix": "Bearer"
                      }
        }

        logging.info("Updating auth for subscriber '{}' with client id '{}'".format(subscriber_id, auth["client_id"]))
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def delete_subscribers(self, client_id):
        path = "/csp/gateway/po/api/v1/subscribers"

        parameters = { "clientID": client_id }

        logging.info("Getting all subscribers for client id '{}'".format(client_id))
        r = self.session.get(path, allow_redirects=True, params=parameters)
        resp = json.loads(r.text)
        sub_ids = [ s["subscriberID"] for s in resp ]
        for sub_id in sub_ids:
            logging.info("Deleting subscriber '{}'".format(sub_id))
            d = self.session.delete(path, allow_redirects=True, params={ "subscriber": sub_id })
            logging.info(d.text)


    def create_subscription(self, subscriber_id, client_id, endpoint, message_type):
        path = "/csp/gateway/po/api/v1/subscribers/{}/subscriptions".format(subscriber_id)

        payload = { "clientID": client_id,
                    "endpoint": endpoint,
                    "messageType": message_type
        }

        logging.info("Create subscription for client id '{}', for message type '{}' to endpoint '{}'".format(client_id, message_type, endpoint))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        resp = json.loads(r.text)

        if r.status_code != 201:
            logging.error("Error creating subsrciption")
            raise RuntimeError(resp)

        return resp["subscriptionID"]


    def create_identity_subscriptions(self, template_file, subscriber_id,
                                      client_id, endpoint, manifest_type):
        templateLoader = jinja2.FileSystemLoader(searchpath="./")
        templateEnv = jinja2.Environment(loader=templateLoader)
        template = templateEnv.get_template(template_file)
        message_types = yaml.safe_load(template.render())["message_types"]

        logging.info("Create Identity subscriptions")
        for mt in message_types:
            if message_types[mt][manifest_type]:
                self.create_subscription(subscriber_id, client_id, endpoint + mt, mt)


    def create_temp_oauth_app(self, env, org_id):
        path = "/csp/gateway/am/api/orgs/{}/oauth-apps".format(org_id)
        payload = {
           "displayName": env + " TEMP BOOTSTRAPPING CLIENT",
           "description": "Temporary OAuth app for bootstrapping purposes ",
           "grantTypes": [
               "client_credentials",
               "password"
           ],
           "accessTokenTTL": 1800,
           "refreshTokenTTL": 7776000,
           "allowedScopes": {
               "allRoles": "false",
               "organizationScopes": {
                   "allRoles": "true",
                   "roleNames": [],
                   "permissions": [],
                   "allPermissions": "true",
                   "keptInToken": [
                       "ROLES"
                   ]
               },
               "servicesScopes": [],
               "generalScopes": []
           },
           "organizationId": org_id,
           "maxGroupsInIdToken": 20
        }

        logging.info("Creating temporary OAuth app for bootstrapping")
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        return { "client_id": resp["clientId"], "client_secret": resp["clientSecret"] }


    def add_service_client_to_org(self, org_id, client_id, org_roles, service_roles, service_definition_id=None):
        path = "/csp/gateway/am/api/orgs/{}/clients".format(org_id)

        payload = {
           "ids": [
               client_id
           ],
           "organizationRoles": list(map(lambda x: { "name": x }, org_roles)),
        }

        if service_definition_id:
            payload.update({ "serviceRoles": [ { "serviceDefinitionId": service_definition_id,
                                                 "serviceRoles": list(map(lambda x: { "name": x }, service_roles)) } ] })


        logging.info("Adding service client '{}' to organization '{}'".format(client_id, org_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def get_group(self, org_id, name):
        path = "/csp/gateway/am/api/orgs/{}/groups".format(org_id)
        logging.info("Getting group '{}' in org '{}'".format(name, org_id))
        r = self.session.get(path, allow_redirects=True)
        resp = json.loads(r.text)
        for group in resp["results"]:
            if name == group["displayName"]:
                return group["id"]

        logging.info("Group '{}' was not found.".format(name))
        return None


    def create_group(self, org_id, name, description):
        path = "/csp/gateway/am/api/orgs/{}/groups".format(org_id)
        payload = {
                    "name": name,
                    "description": description
        }

        logging.info("Creating group '{}'".format(name))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        return json.loads(r.text)["id"]


    def update_group_roles(self, org_id, service_definition_id, service_instance_id, group_id, org_roles_to_add, service_roles_to_add):
        path = "/csp/gateway/am/api/orgs/{}/groups/{}/roles".format(org_id, group_id)
        payload = {
           "organizationRoles": {
               "rolesToAdd": list(map(lambda x: { "resource": "instance:" + service_instance_id, "name": x }, org_roles_to_add))
           },
           "serviceRoles": [
             {
               "serviceDefinitionId": service_definition_id,
               "rolesToAdd": list(map(lambda x: { "resource": "instance:" + service_instance_id, "name": x }, service_roles_to_add)),
               # Makes sure no roles having resource "null" exist.
               "rolesToRemove": list(map(lambda x: { "resource": "null", "name": x }, service_roles_to_add))
             }
           ],
           "notifyUsersInGroups": "false"
        }

        logging.info("Update group roles for group '{}'".format(group_id))
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def create_org(self, display_name, create_service_org=False):
        curr_user_orgs = self.__get_user_orgs(self.get_user_info()["sub"])
        for org in curr_user_orgs:
            if org["displayName"] == display_name:
                raise Exception("Organization with name '{}' already exists.".format(display_name))

        if create_service_org:
            path = "/csp/gateway/am/api/orgs/service"
            resp_key = "id"
        else:
            path = "/csp/gateway/am/api/orgs"
            resp_key = "refLink"

        payload = { "displayName": display_name}

        logging.info("Creating '{}' organization".format(display_name))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        if not r.ok:
            raise Exception("Creation failed")
        resp = json.loads(r.text)
        # Returns confer_org
        return resp[resp_key].split('/')[-1]


    def create_service_invitation(self, service_definition_id, service_instance_id):
        path = "/csp/gateway/slc/api/service-invitations"
        payload = {
            "serviceDefinitionLink": "/csp/gateway/slc/api/definitions/external/{}".format(service_definition_id),
            "numberOfInvitationsToGenerate": 1,
            "isTosPreSigned": "true",
            "isFundMandatory": "false",
            "serviceInstanceIds": [ service_instance_id ]
        }

        logging.info("Creating service invitation for definition '{}' and instance '{}'".format(service_definition_id, service_instance_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        # Returns service_invitation_id
        return resp['refLinks'][0].split('/')[-1]


    def redeem_service_invitation(self, service_invitation_id, org_id):
        path = "/csp/gateway/slc/api/service-invitations/{}".format(service_invitation_id)
        payload = {
            "invitationAction": "REDEEM",
            "orgLink": "/csp/gateway/am/api/orgs/{}".format(org_id)
        }

        logging.info("Redeeming invitation")
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)


    def create_service_instance(self, service_definition_id, display_name, service_url):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances".format(service_definition_id)

        payload = {
            "displayName": display_name,
            "url": service_url
        }

        logging.info("Creating '{}' service instance".format(display_name))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
        resp = json.loads(r.text)
        # Return the service instance id
        return resp['id']


    def get_service_instance_id(self, service_definition_id, display_name):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances".format(service_definition_id)

        logging.info("Getting service instance ID for definition '{}', with name '{}'".format(service_definition_id, display_name))
        all_instances = []
        more_pages = True
        while more_pages:
            r = self.session.get(path, allow_redirects=True)
            if not r.ok:
                logging.critical(r.text)
                raise Exception("Instances could not be retrieved from CSP.")

            r_json = json.loads(r.text)
            all_instances += r_json["results"]
            if r_json["nextLink"]:
                path = r_json["nextLink"]
            else:
                more_pages = False

        for inst in all_instances:
            if inst["displayName"] == display_name:
                return inst["id"]

        logging.info("Instance was not found")
        return None


    def get_service_instance(self, service_definition_id, service_instance_id):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances/{}".format(service_definition_id, service_instance_id)
        logging.info("Getting service instance '{}' for service definition '{}'".format(service_instance_id, service_definition_id))
        r = self.session.get(path, allow_redirects=True)
        return json.loads(r.text)


    def patch_service_instance(self, service_definition_id, service_instance_id, display_name):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances/{}".format(service_definition_id, service_instance_id)
        payload = { "displayName": display_name }
        logging.info("Patching service instance '{}' name to '{}' for service definition '{}'".format(service_instance_id, display_name, service_definition_id))
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info("Status code: {}".format(r.status_code))


    def delete_service_instance(self, service_instance_id, service_definition_id):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances/{}".format(service_definition_id, service_instance_id)
        logging.info("Deleting service instance '{}' for service definition '{}'".format(service_instance_id, service_definition_id))
        r = self.session.delete(path, allow_redirects=True)
        logging.info("Status code: {}".format(r.status_code))


    def assign_org_to_service_instance(self, service_definition_id, org, service_instance_id):
        path = "/csp/gateway/slc/api/definitions/external/{}/service-instances/{}/orgs".format(service_definition_id, service_instance_id)
        payload = { "orgIdsToAdd": [ org ] }

        logging.info("Assigning '{}' organization to service instance '{}'".format(org, service_instance_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info("Status code: {}".format(r.status_code))


    def make_org_operator(self, service_definition_id, org_id):
        path = "/csp/gateway/slc/api/definitions/external/{}/operator-orgs".format(service_definition_id)
        payload = { "orgId": org_id }

        logging.info("Making '{}' organization an operator for '{}'".format(org_id, service_definition_id))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info("Status code: {}".format(r.status_code))


    # Execute with CSR auth, otherwise OPERATOR roles won't be seen
    def get_service_roles(self, service_definition_id):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles".format(service_definition_id)
        parameters = { "includeOperatorRoles": "true" }
        r = self.session.get(path, allow_redirects=True, params=parameters)

        if r.status_code != 200:
            logging.error("Failed to get service roles for '{}'".format(service_definition_id))
            raise RuntimeError(r.text)
        else:
            return json.loads(r.text)["results"]


    def create_service_role(self, service_definition_id, payload):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles".format(service_definition_id)
        logging.info("Creating service role '{}'".format(payload["name"]))
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))

        if r.status_code != 200:
            logging.error("Could not create service role '{}'".format(payload["name"]))
            raise RuntimeError(r.text)
        else:
            logging.info(r.text)


    def patch_service_role(self, service_definition_id, payload):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles/{}".format(service_definition_id, payload["name"])
        logging.info("Patching service role '{}'".format(payload["name"]))
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))

        if r.status_code != 200:
            logging.error("Could not patch service role '{}'".format(payload["name"]))
            raise RuntimeError(r.text)
        else:
            logging.info(r.text)


    def get_service_role(self, service_definition_id, payload):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles/{}".format(service_definition_id, payload["name"])
        logging.info("Get service role '{}'".format(payload["name"]))
        r = self.session.get(path, allow_redirects=True)

        if r.status_code != 200:
            logging.error("Could not get service role '{}'".format(payload["name"]))
            raise RuntimeError(r.text)
        else:
            logging.info(r.text)


    def delete_service_role(self, service_definition_id, name):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles/{}".format(service_definition_id, name)
        logging.info("Deleting service role '{}'".format(name))
        r = self.session.delete(path, allow_redirects=True)

        if r.status_code != 200:
            logging.error("Could not delete service role '{}'".format(name))
            raise RuntimeError(r.text)
        else:
            logging.info("Status code '{}'".format(r.status_code))


    def update_service_definition_roles(self, service_definition_id, definition_template_file, extra_template_file, manifest_type, service_ticker):
        templateLoader = jinja2.FileSystemLoader(searchpath="./")
        templateEnv = jinja2.Environment(loader=templateLoader)
        template = templateEnv.get_template(definition_template_file)
        definition = yaml.safe_load(template.render())
        template_roles = definition["service-roles"]

        extra_template = templateEnv.get_template(extra_template_file)
        extra_definition = yaml.safe_load(extra_template.render({ "service_ticker": service_ticker }))
        template_roles += extra_definition["entities"]["roles"]
        template_permissions = extra_definition["entities"]["permissions"]

        existing_roles = self.get_service_roles(service_definition_id)
        existing_permissions = self.get_permissions(service_definition_id)

        # Update deletion lists
        permissions_to_delete = []
        for ep in existing_permissions:
            for tp in template_permissions:
                ep_name_no_prefix = ':'.join(ep['name'].split(':')[1:])
                if ep_name_no_prefix == tp['name'] and (manifest_type == 'gov' and not tp['gov']):
                    permissions_to_delete.append(ep['name'])

        roles_to_delete = []
        for er in existing_roles:
            delete = True
            for tr in template_roles:
                if er['name'] == tr['name']:
                    delete = False
                    if manifest_type == 'gov' and not tr['gov']:
                        delete = True
            if delete:
                roles_to_delete.append(er['name'])

        # Create/patch permissions
        for tp in template_permissions:
            create = True
            for ep in existing_permissions:
                ep_name_no_prefix = ':'.join(ep['name'].split(':')[1:])
                if (manifest_type == 'gov' and not tp['gov']):
                    create = False
                elif ep_name_no_prefix == tp['name']:
                    create = False
            if create:
                self.create_permission(service_definition_id,
                                       tp['name'],
                                       tp['displayName'],
                                       tp['description'],
                                       tp['customRoleSupport'])

        for tp in template_permissions:
            for ep in existing_permissions:
                if ep['name'] == tp['name']:
                    if ep['displayName'] != tp['displayName'] or ep['description'] != tp['description'] or ep['customRoleSupport'] != tp['customRoleSupport']:
                        self.patch_permission(service_definition_id, tp['name'],
                                              tp['displayName'],
                                              tp['description'],
                                              tp['customRoleSupport'])

        # Update roles creation/patching lists
        roles_to_create = []
        for tr in template_roles:
            create = True
            for er in existing_roles:
                if (manifest_type == 'gov' and not tr['gov']):
                    create = False
                elif er['name'] == tr['name']:
                    create = False
            if create:
                tr.pop('gov')
                roles_to_create.append(tr)

        roles_to_patch = []
        for er in existing_roles:
            for tr in template_roles:
                if er['name'] == tr['name']:
                    if er['displayName'] != tr['display-name'] or er['type'] != tr['type'] or er['visible'] == tr['hidden'] or er['description'] != tr['description']:
                        roles_to_patch.append(tr)

                    if "permissions" in tr:
                        s1 = set(er["permissions"])
                        # Leave the permission that contains the role name always.
                        s1.remove(f"external/{service_definition_id}/{er['name']}")
                        s2 = set(tr["permissions"])
                        to_rm = s1 - s2
                        to_add = s2 - s1
                        if s1 or s2:
                            self.patch_service_role_permissions(service_definition_id, tr['name'], to_add, to_rm)

        roles_plan = { "patch": roles_to_patch, "delete": roles_to_delete, "create": roles_to_create }

        for k in roles_plan.keys():
            if k == "delete":
                for d in roles_plan[k]:
                    self.delete_service_role(service_definition_id, d)
            else:
                for p in roles_plan[k]:
                    p["displayName"] = p.pop("display-name")
                    if p["hidden"] is True:
                        p["visible"] = "false"
                    else:
                        p["visible"] = "true"
                    p.pop("hidden")

                    if k == "create":
                        self.create_service_role(service_definition_id, p)
                    elif k == "patch":
                        self.patch_service_role(service_definition_id, p)

        for pd in permissions_to_delete:
            self.delete_permission(service_definition_id, pd)

        existing_roles = self.get_service_roles(service_definition_id)
        logging.info(f"Printing existing service roles for '{service_definition_id}'")
        logging.info(json.dumps(existing_roles, indent=2))


    def patch_service_role_permissions(self, service_definition_id, role_name, to_add, to_rm):
        path = f"/csp/gateway/iam-roles-mgmt/api/services/{service_definition_id}/roles/{role_name}/permissions"
        to_add = list(to_add)
        to_rm = list(to_rm)
        logging.info(f"Patch permissions for service role '{role_name}'. Add: {to_add}, Remove: {to_rm}")
        payload = {
                      "permissionsToRemove": to_rm,
                      "permissionsToAdd": to_add
        }

        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def create_permission(self, service_definition_id, name, display_name, description, custom_role_support=True):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/permissions".format(service_definition_id)
        logging.info("Creating permission '{}' in '{}'".format(name, service_definition_id))
        payload = {
                    "name": name,
                    "displayName": display_name,
                    "description": description,
                    "customRoleSupport": custom_role_support
        }

        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))

        if not r.ok:
            if r.status_code == 409:
                logging.info(f"Permission '{name}' already exists.")
            else:
                logging.error("Could not create permission {}: {}/{}".format(name, r.status_code, r.text))


    def delete_permission(self, service_definition_id, name):
        path = f"/csp/gateway/iam-roles-mgmt/api/services/{service_definition_id}/permissions/{name}"
        logging.info("Delete permission '{}' in '{}'".format(name, service_definition_id))
        r = self.session.delete(path, allow_redirects=True)
        logging.info(r.text)


    def patch_permission(self, service_definition_id, name, display_name, description, custom_role_support=True):
        # Name is in format serviceTicker:resourceType:action
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/permissions/{}".format(service_definition_id, name)
        logging.info("Patch permission '{}' in '{}'".format(name, service_definition_id))
        payload = {
                    "displayName": display_name,
                    "description": description,
                    "customRoleSupport": custom_role_support
        }

        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))

        if not r.ok:
            logging.error("Could not patch permission {}: {}/{}".format(name, r.status_code, r.text))


    def create_role(self, service_definition_id, name, display_name, description, visible, type, composable, permissions):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/roles".format(service_definition_id)
        logging.info("Create role '{}' in '{}'".format(name, service_definition_id))
        payload = {
                    "name": name,
                    "displayName": display_name,
                    "description": description,
                    "visible": visible,
                    "type": type,
                    "composable": composable,
                    "permissions": permissions
        }

        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))

        if not r.ok:
            logging.error("Could not create role {}: {}/{}".format(name, r.status_code, r.text))


    def get_permissions(self, service_definition_id):
        path = "/csp/gateway/iam-roles-mgmt/api/services/{}/permissions".format(service_definition_id)
        logging.info("Get permissions for '{}'".format(service_definition_id))

        all_permissions = []
        more_pages = True
        while more_pages:
            r = self.session.get(path, allow_redirects=True)
            if not r.ok:
                raise Exception('Problem retrieving permissions from CSP API {}. Got status code: {}. {}'
                                .format(path, r.status_code, r.text))
            csp_resp_json = json.loads(r.text)
            all_permissions += csp_resp_json['results']
            if "nextLink" in csp_resp_json:
                path = csp_resp_json['nextLink']
            else:
                more_pages = False

        return all_permissions


    def set_service_icon(self, service_definition_id, icon_file_path):
        with open(icon_file_path, 'r') as icon_file:
            encoded_icon = base64.b64encode(icon_file.read().encode('utf-8'))

        path = f"/csp/gateway/slc/api/definitions/external/{service_definition_id}/icons"
        logging.info(f"Set icon for '{service_definition_id}'")

        payload = { "svg": encoded_icon.decode('utf-8') }
        r = self.session.post(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)


    def patch_service_urls(self, service_definition_id, request_access_url, pricing_info_url):
        path = f"/csp/gateway/slc/api/definitions/external/{service_definition_id}/service-urls"
        logging.info(f"Patch service URLs for '{service_definition_id}'")

        payload = { "requestAccess": request_access_url, "pricingInformation": pricing_info_url }
        r = self.session.patch(path, allow_redirects=True, data=json.dumps(payload))
        logging.info(r.text)
