import logging

from kubernetes import client, config


class KubernetesClient:
    def __init__(self):
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException as incluster_e:
            logging.info("Incluster configuration unavailable. Falling back to kube-config...")
            if not config.load_kube_config():
                raise config.config_exception.ConfigException

        self.core_api = client.CoreV1Api()


    def create_secret(self, secret_name, secret_data, namespace):
        secret_body = client.V1Secret(api_version="v1",
                                      kind="Secret",
                                      metadata=client.V1ObjectMeta(name=secret_name,
                                                                   namespace=namespace),
                                      string_data=secret_data,
                                      type="Opaque")
        self.core_api.create_namespaced_secret(namespace, body=secret_body)
        logging.info("Created Kubernetes secret '{}' in '{}'".format(secret_name, namespace))


    def get_secret(self, secret_name, namespace):
        try:
            return self.core_api.read_namespaced_secret(secret_name, namespace).to_dict()['data']
        except:
            return None


    def delete_secret(self, secret_name, namespace):
        try:
            self.core_api.delete_namespaced_secret(secret_name, namespace)
            logging.info("Deleted Kubernetes secret '{}' in '{}'".format(secret_name, namespace))
        except client.exceptions.ApiException as e:
            if e.reason != "Not Found":
                raise e
            else:
                logging.info("Kubernetes secret '{}' does not exist in '{}'. Nothing to delete.".format(secret_name, namespace))
                return


    def create_config_map(self, name, config_map_dict, namespace):
        body = client.V1ConfigMap(kind="ConfigMap",
                              metadata={ "name": name,
                                         "namespace": namespace },
                              data=config_map_dict)

        self.core_api.create_namespaced_config_map(namespace, body)
        logging.info("Created Kubernetes ConfigMap '{}' in '{}'".format(name, namespace))


    def delete_config_map(self, config_map_name, namespace):
        try:
            self.core_api.delete_namespaced_config_map(config_map_name, namespace)
            logging.info("Deleted Kubernetes config map '{}' in '{}'".format(config_map_name, namespace))
        except client.exceptions.ApiException as e:
            if e.reason != "Not Found":
                raise e
            else:
                logging.info("Kubernetes config map '{}' does not exist in '{}'. Nothing to delete.".format(config_map_name, namespace))
                return


    def read_config_map(self, name, namespace):
        try:
            return self.core_api.read_namespaced_config_map(name, namespace).to_dict()
        except client.exceptions.ApiException as e:
            if e.reason == "Not Found":
                logging.info("ConfigMap '{}' was not found in namespace '{}'".format(name, namespace))
            else:
                print(e)

            return None


    def get_key_from_config_map(self, key_name, config_map_name, namespace):
        cm = self.get_config_map(config_map_name, namespace)
        if cm:
            try:
                return cm[key_name]
            except KeyError as k:
                logging.info("Key '{}' not found in ConfigMap '{}' in '{}'".format(key_name, config_map_name, namespace))

        return None


    def get_config_map(self, config_map_name, namespace):
        cm = self.read_config_map(config_map_name, namespace)
        if cm:
            return cm["data"]
        return None


    def patch_config_map(self, name, config_map_dict, namespace):
        body = client.V1ConfigMap(kind="ConfigMap",
                                  metadata={ "name": name,
                                             "namespace": namespace },
                                  data=config_map_dict)

        self.core_api.patch_namespaced_config_map(name, namespace, body)
        logging.info("Patched ConfigMap '{}' in '{}'".format(name, namespace))
