import json
import logging
import base64

import requests


class RequestsSession(requests.Session):
    """Wrapper class for requests.Session to store base URL and headers"""

    def __init__(self, base_url=None, **kwargs):
        super(RequestsSession, self).__init__(**kwargs)
        self.base_url = base_url


    def get(self, path, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).get(full_url, **kwargs)


    def delete(self, path, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).delete(full_url, **kwargs)


    def post(self, path, data=None, json=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).post(full_url, data, json, **kwargs)


    def put(self, path, data=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).put(full_url, data, **kwargs)


    def patch(self, path, data=None, **kwargs):
        full_url = '{}{}'.format(self.base_url, path)
        return super(RequestsSession, self).patch(full_url, data, **kwargs)


class CSPSession(RequestsSession):
    def __init__(self, base_url, **kwargs):
        super(CSPSession, self).__init__(base_url=base_url)
        try:
            self.authenticate(**kwargs)
        except:
            logging.critical("CSP access token could not be acquired")
            raise
        self.headers.update({ "Content-Type": "application/json" })

    # check kwargs all keys
    def authenticate(self, **kwargs):
        if "grant_type" not in kwargs:
            raise Exception("'grant_type' not provided")

        if kwargs["grant_type"] == "api_token":
            if "api_token" not in kwargs:
                raise Exception("'api_token' parameter is required when using grant_type 'api_token'")
            self.access_token = self.__get_access_token_from_api_token(api_token=kwargs["api_token"])
        elif kwargs["grant_type"] == "password":
            if "client_id" not in kwargs or "client_secret" not in kwargs or "username" not in kwargs or "password" not in kwargs:
                raise Exception("'client_id', 'client_secret', 'user' and 'password' parameters are required when using grant_type 'password'")
            self.access_token =  self.__get_access_token_from_client_creds(**kwargs)
        elif kwargs["grant_type"] == "client_credentials":
            if "client_id" not in kwargs or "client_secret" not in kwargs:
                raise Exception("'client_id' and 'client_secret' parameters are required when using grant_type 'client_credentials'")
            self.access_token =  self.__get_access_token_from_client_creds(**kwargs)
        else:
            raise Exception("Invalid grant_type")
        self.headers.update({ "Authorization": "Bearer " + self.access_token })


    def __get_access_token_from_api_token(self, api_token):
        logging.info("Acquiring access token from API token.")
        r = super().post("/csp/gateway/am/api/auth/api-tokens/authorize",
                        headers={ "Content-Type": "application/x-www-form-urlencoded" },
                        data={ "api_token": api_token },
                        allow_redirects=True)

        resp = json.loads(r.text)
        if "access_token" not in resp:
            logging.info(resp)
            raise Exception("CSP response does not contain an access token")

        return resp["access_token"]


    def __get_access_token_from_client_creds(self, **kwargs):
        basic_creds = base64.b64encode("{}:{}".format(kwargs["client_id"], kwargs["client_secret"]).encode('utf-8')).decode('utf-8')
        logging.info("Acquiring access token from client credentials.")
        r = super().post("/csp/gateway/am/api/auth/token",
                         headers={ "Content-Type": "application/x-www-form-urlencoded",
                                   "Authorization": "Basic " + basic_creds },
                         data=kwargs)

        resp = json.loads(r.text)
        if "access_token" not in resp:
            logging.info(resp)
            raise Exception("CSP response does not contain an access token")

        return resp["access_token"]
